//==========================================================================
//
//      net/httpd.c
//
//      Stand-alone HTTPD support for RedBoot
//
//==========================================================================
//==========================================================================
//#####DESCRIPTIONBEGIN####
// Author(s):    Sam Huang
// Date:         2005-10-16
// Purpose:
//         Perform httpd service in RedBoot.
// Description:
//         This code httpd application and only running for RedBoot.
//
//#####DESCRIPTIONEND####
//==========================================================================

#include <common.h>
#include <command.h>
#include "netif.h"
#include <malloc.h>

#if 0
#define httpd_debug
#endif

#define false -1
#define true 0

#define ISO_nl       0x0a
#define ISO_cr       0x0d
#define PAGE_SIZE    1024

#define HTTP_NOGET   0
#define HTTP_FILE    1
#define HTTP_TEXT    2
#define HTTP_FUNC    3
#define HTTP_END     4
#define HTTP_UPLOAD  5

static tcp_socket_t httpd_sock;
static int        httpd_state;
static int        httpd_rxlen = 0;
//static unsigned char httpd_rxbuf[2048];
static unsigned char httpd_rxbuf[4096];
static unsigned char httpd_txbuf[PAGE_SIZE];
static int	  httpd_txlen = 0;

/* for file upload */
struct httpd_hs {
	unsigned char  state;
	unsigned short count;
	char *dataptr;
};
struct httpd_hs http_stat;
struct httpd_hs *hs=&http_stat;

static char ul_boundary[64];
static int ul_len;
static int ul_cnt;
static int ul_offset;

static unsigned int DstAddr;
static unsigned int DstOffset;

static unsigned char bUploadIsOK = 0;

#define FLASH_IMAGE_ADDR  CFG_RUNTIME_START_ADDR

// for image header check
#define IMG_MAX_DEVNAME 32
#define MAX_SIGNATURE   32

typedef struct _imgblock imgblock_t;
struct _imgblock
{
	unsigned long magic;
	unsigned long size;
	unsigned long offset;
	char		devname[IMG_MAX_DEVNAME];
	unsigned char digest[16];
} __attribute__((packed));

typedef struct _imghdr2 imghdr2_t;
struct _imghdr2
{
	char signature[MAX_SIGNATURE];
	unsigned long magic; /*should be IMG_V2_MAGIC_NO in little endian */
} __attribute__ ((packed));
#define DBYTES(data, size)

extern void *flash_start,*flash_end;

#define	INDEX_PAGE "<html><body>\n"\
"<center><br><br><h2>Emergency Web Server</h2>\n"\
"<form action=up.htm method=post encType=multipart/form-data>\n"\
"File <input type=file size=35 name=files><input type=submit value=upload></form><br>\n"\
"</center></body></html>\n"


#define WAIT_PAGE   "<html><head>\n"\
"<META HTTP-EQUIV=Content-Type CONTENT='no-cache'>\n"\
"<META HTTP-EQUIV=Content-Type CONTENT='text/html; charset=iso-8859-1'>\n"\
"</head><script>var countdown=60;function nev(){countdown--;document.formWaitInfo.WaitInfo.value=countdown;\n"\
"if(countdown < 1) self.location.href='OK.html'; else setTimeout('nev()',1000);}</script>\n"\
"<body onload='nev()'><form name='formWaitInfo'><center><br><br><h2>Updating File </h2><hr><br><br><FONT SIZE=4>\n"\
"<font color=red>Don't Power Down.</font><br><br> Please wait for <input type='Text' readonly name='WaitInfo' size='3' style='border-width:0; background-color=#FFFFFF; color:#FF0000; text-align:center'> seconds...</center>\n"\
"</form></body></html>\n"

#define OK_PAGE "<html><body><br><br>\n"\
"<center><h2>Update File Successfully</h2><hr><br><br><FONT SIZE=4>Please restart device...</center>\n"\
"</body></html>\n"

#define diag_printf   printf
#define diag_sprintf  sprintf

#define FREE_MEM_ADDR	CFG_LOAD_ADDR

#if 0
RedBoot_cmd("httpd","Start httpd",
	"",
		do_httpd
);

static int httpd_flash_write(char *src_addr,char *dst_addr,int size)
{
	void *err_addr;
	int stat;

	if ((stat = flash_erase(dst_addr, size, (void **)&err_addr)) != 0) {
	diag_printf("httpd_flash_write: Error erasing at %p: %s\n",err_addr, flash_errmsg(stat));
	return 0;
	} else {
	if ((stat = FLASH_PROGRAM( dst_addr, src_addr, size, (void **)&err_addr)) != 0) {
		diag_printf("httpd_flash_write: Error writing at %p: %s\n",err_addr, flash_errmsg(stat));
		return 0;
		}
	}
	return 1;
}
#endif

static void do_httpd(int argc, char *argv[])
{
	diag_printf("Start httpd service...\n");
	httpd_init();
}

static void httpd_init_hs(void)
{
	//memset(hs, 0, sizeof(*hs));
	memset(&http_stat, 0, sizeof(http_stat));
}

static void httpd_printf(char *m)
{
	char c;
	if (httpd_txlen > (PAGE_SIZE-32)) return ;
	while (c=*m++)
	{
		httpd_txbuf[httpd_txlen++]=c;
		if (c=='\r')
			httpd_txbuf[httpd_txlen++]='\r';
	}
}

static void httpd_page_init(void)
{
	httpd_txlen=diag_sprintf(httpd_txbuf,"<HTML><BODY>\n");
}

static void httpd_page_end(int rc,int f)
{
	httpd_txlen+=diag_sprintf(httpd_txbuf+httpd_txlen, "\n<!-- rc=%d -->%s</BODY></HTML>\n", rc, f? "<a href=index.htm>Back</a>" : "" );
}

static int power(int base, int power)
{
	int sum = base;

	if (power == 1)
		return sum;
	else
		if (power == 0)
			return 1;

	for(; power>1; power--)
	{
		sum *= base;
	}
	return sum;
}

static void httpd_scanf(char *s,int *len)
{
	char *tmp;
	int i,j,k;

	*len=i=j=0;
	tmp = s;
	for (j=0 ; /*((*tmp != 0x0d)&&(*(tmp+1)!=0x0a)) ||*/ *tmp!=0x0 ; j++ )
	{
		if (*tmp < 0x30 || *tmp > 0x39)
		{
			return;
		}
		tmp++;
	}

	if (j == 0) return;

	k = j;
	tmp = s;
	while ( i < j )
	{
		k--;
		*len += (tmp[i]-0x30)* power(10,k);
		i++;
	}
}

#if 0
static int httpd_assemble_img(char *image, int size, unsigned int dst_addr,unsigned int *dst_offset)
{
	int offset = sizeof(imghdr2_t);
	imgblock_t *block = 0;
		unsigned int dst_tmp = dst_addr;
	unsigned int src_tmp = 0;

	while (offset < size)
	{
		block = (imgblock_t *)&image[offset];

		diag_printf("Image header (0x%08x):\n", offset);
		diag_printf("  magic  : 0x%08x\n", block->magic);
		diag_printf("  size   : %d (0x%x)\n", block->size, block->size);
		diag_printf("  offset : 0x%08x\n", block->offset);
		diag_printf("  devname: \'%s\'\n", block->devname);

		/* move to next block */
		offset += sizeof(imgblock_t);

		src_tmp = (unsigned int)(block);
		src_tmp += sizeof(imgblock_t));

		diag_printf("src_tmp=0x%x\n",src_tmp);

		memcpy(dst_tmp, src_tmp, block->size);
		*dst_offset += block->size;
		dst_tmp += block->size;

		offset += block->size;

		diag_printf("Advance to next block : offset=%d(0x%x), size=%d(0x%x)\n", offset, offset, size, size);
	}
	if (offset == size) return 0;

	diag_printf("illegal block found at offset %d (0x%x)!\n", offset, offset);
	diag_printf("  offset (%d) : \n", offset);

	return -1;
}
#endif

static int httpd_fileupload(void)
{
	char buf[256];
	char *line,*end;
	int j;

	unsigned int load_addr = FREE_MEM_ADDR;

#ifdef httpd_debug
printf("Loading file in memory(0x%x).\n", load_addr);
printf("[%s:%d]\n\t http_stat.count=%d \n", __FUNCTION__, __LINE__, http_stat.count);
#endif
	if(http_stat.count < 4)
	{
		line=end=httpd_rxbuf;

		// found content-length , content-type, and data
		while( end < (httpd_rxbuf+httpd_rxlen) && http_stat.count<5 )
		{
			if(!(*end == ISO_cr && *(end+1) == ISO_nl))
			{
				end++;
				continue;
			}
			*end='\0';
			// state 3
			if(http_stat.count == 3)
			{
				// search for data
				if (line==end) //hs->count data
				{
					end+=2;
					ul_offset=((unsigned int)httpd_rxbuf+httpd_rxlen)-(unsigned int)end;
					memcpy((char *)load_addr,(char *)end, ul_offset);
#ifdef httpd_debug
			diag_printf("count %d data %d\n", http_stat.count, ul_offset) ;
#endif
					http_stat.count++;
					break;
				}
			}
			else
				// state 2
				if(http_stat.count == 2)
				{
					// search for boundary
					for(j=0;j<60;j++)
						if (line[j]!='-') break;

					if (!strncmp(line+j,ul_boundary,strlen(ul_boundary)))
					{
#ifdef httpd_debug
						diag_printf("count %d boundary ", http_stat.count);
#endif
						http_stat.count++;
						ul_cnt=((unsigned int)httpd_rxbuf+httpd_rxlen)-(unsigned int)line ;
#ifdef httpd_debug
						diag_printf("httpd_rxlen=%d ul_cnt=%d\n", httpd_rxlen, ul_cnt) ;
#endif
					}
				}
				else
				if(http_stat.count <= 1)
				{ // state 0,1
					if(http_stat.count == 1)
						httpd_printf("<pre>");

					for(j=0;(line+j)<end;j++)
						buf[j]=line[j]|0x20;

					buf[j]='\0';

					if (!strncmp("content-length",buf,14))
					{
						int flash_size;
#ifdef httpd_debug
						diag_printf("before scanf..%s\n",buf+16);
						diag_printf("buf[16]=0x%x,buf[17]=0x%x,buf[18]=0x%x,buf[19]=0x%x,buf[20]=0x%x,buf[21]=0x%x,buf[22]=0x%x\n",*(buf+16),*(buf+17),*(buf+18),*(buf+19),*(buf+20),*(buf+21),*(buf+22));
#endif
						httpd_scanf(buf+16,&ul_len);
#ifdef httpd_debug
						diag_printf("count %d content-length=%d\n", http_stat.count, ul_len);
#endif
		//samh TODO:
#if 0
						// check image type according to flash size.
						flash_size = (int)(flash_end - flash_start + 0x10000 + 1);
						if (ul_len > flash_size)
						{
							diag_printf("wrong image..\n");
							httpd_printf("</pre>");
							httpd_printf("<FONT SIZE=4>Wrong image! Back to select again...<HR>");
							httpd_page_end(0,1);
							return 0;
						}
#endif
						http_stat.count++;
					}
					else
					if (!strncmp("content-type",buf,12))
					{
#ifdef httpd_debug
						diag_printf("content-type: count=%d\n", http_stat.count);
#endif
						for(j=48;j<80;j++)
						{
							if (line[j]!='-')
							{
#ifdef httpd_debug
								diag_printf("count %d content-type boundary=%s\n", http_stat.count, line+j);
#endif
								strcpy(ul_boundary, line+j);
								http_stat.count++;
								break;
							}
						}
					}
				}
			end+=2;
			line=end;
		} //while
		return 1;
	}
	else
		if(http_stat.count == 4)
		{
			if ( (ul_cnt+httpd_rxlen) >= ul_len)
			{
#ifdef httpd_debug
				diag_printf("count %d total rx=%d\n", http_stat.count, ul_cnt+httpd_rxlen);
#endif
				int bl=strlen(ul_boundary);
				j = httpd_rxlen - bl - 6;
				for( ; j > httpd_rxlen - bl - 40 ; j-- )
				{
					if (httpd_rxbuf[j]!='-')
					{
						j--;
						if (httpd_rxbuf[j]!='\x0d' || httpd_rxbuf[j+1]!='\x0a') break;
						httpd_rxlen=j; // only copy those effective data
						break;
					}
				}
				http_stat.count++;
			}
			memcpy((char *)(load_addr+ul_offset),(char *) httpd_rxbuf, httpd_rxlen);
			ul_cnt+=httpd_rxlen ;
			ul_offset+= httpd_rxlen ;

			if(http_stat.count == 5)
			{
				//httpd_printf("</pre>");
				//httpd_printf("<FONT SIZE=4>Update file successfully! Please restart device...<HR>");
				//httpd_page_end(0,1);
#ifdef httpd_debug
				diag_printf("rx len=%d (0x%x)\n",ul_offset, ul_offset);
#endif
				DstAddr = load_addr + sizeof(imghdr2_t) + sizeof(imgblock_t);
				DstOffset = ul_offset - sizeof(imghdr2_t) - sizeof(imgblock_t);
				//DstAddr = load_addr;
				//DstOffset = ul_offset;
#ifdef httpd_debug
				diag_printf("DstAddr=0x%x ,DstOffset=0x%x\n",DstAddr,DstOffset);
#endif
				//write image into flash
				//httpd_flash_write(DstAddr, FLASH_IMAGE_ADDR, DstOffset);
				bUploadIsOK = 1;
				//send waiting page
				httpd_txlen = 0;
				httpd_printf(WAIT_PAGE);
				diag_printf("Upload file OK!\n");
				return 0;
			}
		}
	return 1;
}

/*
 * Handle incoming HTTP packets.
 */
static int httpd_handler(void)
{
	int rc;

	//int len;
	//httpd_scanf("257807\r\n",&len);
	//diag_printf("len=%d\n",len);

	httpd_rxlen = __tcp_read(&httpd_sock, httpd_rxbuf, sizeof(httpd_rxbuf));
#ifdef httpd_debug
printf("[%s:%d]\n\t httpd_rxlen=%d, http_stat.state=0x%x \n", __FUNCTION__, __LINE__, httpd_rxlen, http_stat.state);
#endif
	//if ((httpd_rxlen > 0) && hs->state == HTTP_NOGET )
	if ((httpd_rxlen > 0) && http_stat.state == HTTP_NOGET )
	{
#if 0//def httpd_debug
		printf("============== Receive Packet Contain ==============\n");
		printf("%s",httpd_rxbuf);
		printf("============== ====================== ==============\n");
#endif
		/* Check for POST */
		if (!strncmp("POST ", &httpd_rxbuf[0], 5))
		{
			//POST
			if (!strncmp("up",&httpd_rxbuf[6],2))
			{
				diag_printf("Post up.htm\n");
				//hs->state=HTTP_UPLOAD ;
				http_stat.state = HTTP_UPLOAD;
				//httpd_page_init();
				goto DO_UPLOAD;
			}
			goto GET_PAGE;
		}

		/* Check for GET */
		if(strncmp("GET ",&httpd_rxbuf[0],4))
		{
			/* If it isn't a GET, we abort the connection. */
			diag_printf("isn't GET,return -1\n");
			return -1;
		}

		/* Check for GET OK page */ //"GET /OK.html****"
		if (!strncmp("OK", &httpd_rxbuf[5], 2))
		{
//GET_OK_PAGE:
			//diag_printf("get OK page.\n");
			httpd_txlen = 0;
			httpd_printf(OK_PAGE);
			return 0;
		}

GET_PAGE:
#ifdef httpd_debug
		diag_printf("Client requests index page\n");
#endif
		httpd_page_init();
		httpd_printf(INDEX_PAGE);
		httpd_page_end(0,0);
		return 0;
	}
	else if (httpd_rxlen <= 0)
	{
		// !!! for 5VT board,Don't remove "printf()" below
		diag_printf("httpd: rx=%d\r\b",httpd_rxlen);
		return 1;
	}

DO_UPLOAD:
	//httpd_page_init();
	printf("upload: rx=%d\r\b",httpd_rxlen);
	rc = httpd_fileupload();

	return rc;
}


static int
httpd_read_resp_nonblock(void)
{
	int rc;

	__tcp_poll();
	if (httpd_sock.state == _CLOSE_WAIT)
	{
		// This connection is breaking
		if (httpd_sock.data_bytes == 0 && httpd_sock.rxcnt == 0)
		{
			diag_printf("databytes&rxcnt = 0\n");
			__tcp_close(&httpd_sock);
			return false;
		}
	}
	if (httpd_sock.state == _CLOSED)
	{
		// The connection is gone
		diag_printf("The connection is gone\n");
		return true;
	}

	if ((rc = httpd_handler()) >= 0)
	{
		/* TODO: temperarily use this method
		 */
		if (rc == 0)
		{
			/* Normally response get request from http client */
#ifdef httpd_debug
			diag_printf("Send close..\n");
#endif
			__tcp_write(&httpd_sock,httpd_txbuf,httpd_txlen);
			__tcp_close(&httpd_sock);
		}
		else
		{
#ifdef httpd_debug
			printf("[%s:%d]\n\t got data.. send ack : rc(%d)\n", __FUNCTION__, __LINE__, rc);
#endif
			/* PS: No need to send TCP_ACK by myself,
			 *		  TCP session will do this...
			 */
			//__tcp_send_ack(&httpd_sock);
		}
		return true;
	}
	else
	{
		/* TODO: I think there could another better way to close connection,
		 *	   temperarily use regular method.
		 */
#ifdef httpd_debug
		diag_printf("httpd_handler() returns false\n");
#endif
		__tcp_close(&httpd_sock);
		return false;
	}
}

void httpd_init(void)
{
#ifdef httpd_debug
	diag_printf("%s: httpd init...\n", __FUNCTION__);
#endif
	__tcp_listen(&httpd_sock,80);
	httpd_state = httpd_sock.state;
}

void httpd_accept(int is_idle)
{
	if (!is_idle) {};
	//if (!is_idle) return;  // Only care about idle case
	//if (!have_net) return;
	__tcp_poll();
	if (httpd_state != httpd_sock.state)
	{
#ifdef httpd_debug
printf("[%s:%d]\n\t httpd_state=%d,http_sock.state=%d \n", __FUNCTION__, __LINE__, httpd_state, httpd_sock.state);
#endif
		// Something has changed
		if (httpd_sock.state == _ESTABLISHED)
		{
			httpd_init_hs();
			// A new connection has arrived
			if (!httpd_read_resp_nonblock())
			{
#ifdef httpd_debug
printf("[%s:%d]\n\t httpd_sock.state == established: fail to httpd_read_resp_nonblock \n", __FUNCTION__, __LINE__);
#endif
			}
		}
		if (httpd_sock.state == _TIME_WAIT)
		{
#ifdef httpd_debug
			diag_printf("tcp reset\n");
#endif
			__tcp_reset(&httpd_sock);
		}
		if (httpd_sock.state == _CLOSED)
		{
			// Get ready for another connection
#ifdef httpd_debug
			diag_printf("httpd re-init\n");
#endif
			httpd_init();
		}
#ifdef httpd_debug
printf("[%s:%d]\n\t httpd_state=%d,http_sock.state=%d \n", __FUNCTION__, __LINE__, httpd_state, httpd_sock.state);
#endif
	}
	else if( httpd_state == _ESTABLISHED && httpd_sock.state == _ESTABLISHED)
	{
		if (!httpd_read_resp_nonblock())
		{
#ifdef httpd_debug
printf("[%s:%d]\n\t httpd_state == established: fail to httpd_read_resp_nonblock \n", __FUNCTION__, __LINE__);
#endif
		}
	}

	if (bUploadIsOK == 1)
	{
		int bRemainder = 0,sector_num = 0;
		unsigned long addr = 0;

		sector_num = DstOffset / 0x10000;
		bRemainder = (DstOffset % 0x10000)? 1:0;

		if (bRemainder) sector_num++;

		addr = simple_strtoul(getenv("runtimestart"), NULL, 16);

		if(!addr)
			addr = FLASH_IMAGE_ADDR;

		printf("%s begin addr=0x%x,end=0x%x\n", addr?"(ENV)":"", addr, addr+(sector_num*0x10000)-1);
		flash_sect_erase((unsigned long)addr,(unsigned long)(addr + (sector_num*0x10000)-1));
		flash_write((char *)DstAddr,(char *)addr,DstOffset);

		bUploadIsOK = 0;
		diag_printf("\nwrite image into flash...OK,dest addr=0x%x\n", addr);
	}
	httpd_state = httpd_sock.state;
}
